# -*- coding: ISO-8859-15 -*-
# Copyright (c) 2004 Nuxeo SARL <http://nuxeo.com>
# Copyright (c) 2004 CGEY <http://cgey.com>
# Copyright (c) 2004 Ministre de L'intrieur (MISILL)
#               <http://www.interieur.gouv.fr/>
# Author: Julien Anguenot <ja@nuxeo.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id: BaseMailFolder.py,v 1.7 2004/05/07 12:18:20 anguenot Exp $

__author__ = "Julien Anguenot <ja@nuxeo.com>"

"""Base Mail Folder class.
 """

from DateTime import DateTime

from Globals import InitializeClass
from AccessControl import ClassSecurityInfo

from Products.CMFCore.CMFCorePermissions import View, ModifyPortalContent

from Products.MisillWorkflowEngine.FlexibleWorkflowedDocument import \
     FlexibleWorkflowedDocument

class BaseMailFolder(FlexibleWorkflowedDocument):
    """Base MailFolder Class
    """

    security = ClassSecurityInfo()

    def __init__(self, id, **kw):
        """BaseMailFolder constructor

        Call the CPSDocument constructor
        and then setup some attributes.
        """
        FlexibleWorkflowedDocument.__init__(self, id, **kw)

        # Keeping track of the mailfolder delay
        self.delay_frozen_value = None
        self.delay_time_closed = 0

    security.declareProtected(ModifyPortalContent, 'setIdentifier')
    def setIdentifier(self, context):
        """Set unique identifier for mailfolders
        """

        field_id = 'identifier'

        data_model = self.getTypeInfo().getDataModel(self,
                                                     context)

        #
        # Compute Mail Folder Identifier
        # This is not the Zope internal id but the mail folder
        # identifier used by the organization.
        #

        portal = self.portal_url.getPortalObject()
        root_mail_folder = self.getRootMailFolder() # skins

        #
        # Setting a specific identifier only of the portal_types not
        # declared below
        #

        if self.portal_type not in ['ParliamentMailFolder',
                                    'ParliamentQuestionMailFolder']:
            root_mail_folder = self.getRootMailFolder() # skins

            base_identifier = str(root_mail_folder.getMailFolderIdentifier())
            zeros = 6 - len(base_identifier)
            base_identifier = "0"*zeros + base_identifier

            global_identifier = getattr(portal, 'Direction', 'CPS') \
                                + '-' \
                                + DateTime().strftime("%Y") \
                                + '-' \
                                + base_identifier
        else:
            global_identifier = self.number

        #
        # Set identifier / commit changes
        #

        data_model.set('identifier', global_identifier)
        data_model._commit()

    security.declarePublic("getDaysBeforeDeadlineExpr")
    def getDaysBeforeDeadlineExpr(self):
        """ Returns how many days before the deadline.

        May be None if no delay.
        """
        today = DateTime()
        creation = self.CreationDate()
        if getattr(self, 'delay'):
            delay = self.delay
            # FIXME schemas
            try:
                days = int(round(DateTime(creation)
                                 +  delay
                                 - today
                                 - self.delay_time_closed))
                c = "-"
                if days <0:
                    c = "+"
                return 'J'+ c + str(days)
            except (ValueError):
                pass
        else:
            return 'NS'

    security.declareProtected(ModifyPortalContent, 'grantOwnerManagerRole')
    def grantOwnerManagerRole(self, context=None):
        """ Grants Owner Manager role

        Owner is not enough in CPS.
        """
        if context is not None:
            context.manage_setLocalRoles(self.Creator(),
                                         ['Owner', 'WorkspaceManager'])
            # Substitutes
            context.getContent().setSubstitutesFor(self.Creator(),
                                                   0,
                                                   'pilot',
                                                   context)

    security.declareProtected(View, 'addContributor')
    def addContributor(self, member_id, context):
        """Add a contributor name.

        All the people modifiying the reference documents.
        The attribut is on the schema.
        Called when reference documents are modified.
        """
        aclu = self.acl_users
        field_id = 'contributors'
        data_model = self.getTypeInfo().getDataModel(self, context)
        current_field = data_model.get(field_id)
        if current_field == '':
            current_field = []
        if member_id not in current_field:
            data_model.set(field_id, current_field + [member_id])
            data_model._check_acls = 0
            data_model._commit(check_perms=0)

    security.declareProtected(View, 'addPilot')
    def addPilot(self, member_id, context):
        """Add a pilot name.

        Every pilots through the mail folder life cycle.
        The attribut is on the schema.
        """
        field_id = 'all_pilots'
        data_model = self.getTypeInfo().getDataModel(self, context)
        current_field = data_model.get(field_id)
        if current_field == '':
            current_field = []
        if member_id not in current_field:
            data_model.set(field_id, current_field + [member_id])
            data_model._commit(check_perms=0)

    ###################################################################
    ###################################################################

    security.declareProtected(ModifyPortalContent, 'freezeDelay')
    def freezeDelay(self):
        """Freeze delay while closing the mailfolder

        Will run again when the mailfolder will be re-open
        """
        today = DateTime()
        creation = self.CreationDate()
        if getattr(self, 'delay'):
            delay = self.delay
            self.delay_frozen_value = int(round(DateTime(creation) \
                                                + delay \
                                                - today
                                                - self.delay_time_closed))

    security.declareProtected(ModifyPortalContent, 'unFreezeDelay')
    def unFreezeDelay(self):
        """Un freeze delay while reopening the mailfolder

        Then the delay run again.
        """
        today = DateTime()
        creation = self.CreationDate()
        if self.delay_frozen_value is not None:
            if getattr(self, 'delay'):
                delay = self.delay
                self.delay_time_closed += int(round(DateTime(creation) + delay - today)) \
                                          - self.delay_frozen_value
            self.delay_frozen_value = None

    ###################################################################
    ###################################################################

InitializeClass(BaseMailFolder)
