# -*- coding: ISO-8859-15 -*-
# Copyright (c) 2004 Nuxeo SARL <http://nuxeo.com>
# Copyright (c) 2004 CGEY <http://cgey.com>
# Copyright (c) 2004 Ministre de L'intrieur (MISILL)
#               <http://www.interieur.gouv.fr/>
# Author : Julien Anguenot <ja@nuxeo.com>

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# $Id: Substitutes.py,v 1.4 2004/07/01 16:47:04 regebro Exp $

__author__ = "Julien Anguenot <mailto:ja@nuxeo.com>"

"""Substitutes

It implements the substitute concepts.
It is used within the assignment process with Mailfolders.

The API is in use within the MisillMailManager Class. It is not in us within
the CPSMailManager product right now since CPSDefault doesn't provide ant
substitutes facilities right now.
"""

from zLOG import LOG, DEBUG, INFO

from Globals import InitializeClass
from AccessControl import ClassSecurityInfo

import ExtensionClass
from Products.CMFCore.CMFCorePermissions import ModifyPortalContent
from Products.CMFCore.utils import getToolByName

class Substitutes(ExtensionClass.Base):
    """Substitutes

    It implements the substitute concepts.
    It is used within the assignment process with Mailfolders.

    The API is in use within the MisillMailManager Class. It is not in us
    within the CPSMailManager product right now since CPSDefault doesn't
    provide ant substitutes facilities right now.
    """

    security = ClassSecurityInfo()

    def __init__(self):
        """Constructor
        """

        #
        # Substitutes lists
        # Contain a structure of this type :
        # It avoids to recompute them at any time
        # Expecially, at display time
        # {0:[{'member_id':['sub1_id', 'sub2_id']},.....]}
        #

        self.pilot_level_substitutes = {}
        self.pre_pilot_level_substitutes = {}
        self.associate_level_substitutes = {}
        self.watcher_level_substitutes = {}

    ###################################################################
    ###################################################################

    security.declareProtected(ModifyPortalContent,  'getSubstitutesFor')
    def getSubstitutesFor(self, member_id):
        """Return the substitutes for a given Member

        It returns a list of member_id/group_id
        """

        substitutes = []

        #
        # This is a group -> no substitutes
        # We fetch the substitutes of *all* members.
        #

        if ':' in member_id:
            all_group_members = self._getGroupMembers(member_id.split(':')[1])
            for gmember_id in all_group_members:
                tmp = []
                tmp = self.getSubstitutesFor(gmember_id)
                for id in tmp:
                    if id not in substitutes:
                        substitutes.append(id)
            return substitutes

        portal = self.portal_url.getPortalObject()
        prop = "substitutes"

        # MISILL substitutes place
        if getattr(portal, 'PersoRoot', None) is not None:
            folder = portal[getattr(portal, 'PersoRoot')].members
            folder = getattr(folder, member_id, None)

            if folder is not None:
                folder = folder.getContent()
            else:
                return []

            if folder.hasProperty(prop):
                substitutes = folder[prop]

        return self.filterExistingUsers(substitutes)

    #############################################################
    #############################################################

    security.declarePublic('isPilotSubstituteHere')
    def isPilotSubstituteHere(self, member_id, level):
        """ Is the given member a substitute at this level ?

        If so returns the member for whom he's been set as a substitute
        """

        res = []

        if self.pilot_level_substitutes.has_key(level) and \
               self.pilot_level_substitutes[level]:
            for pilot in self.pilot_level_substitutes[level].keys():
                if member_id in self.pilot_level_substitutes[level][pilot]:
                    res.append(pilot)
        return self.filterExistingUsers(res)

    security.declarePublic('isPrePilotSubstituteHere')
    def isPrePilotSubstituteHere(self, member_id, level):
        """ Is the given member a substitute at this level ?

        If so returns the member for whom he's been s as a substitute
        """

        res = []

        if self.pre_pilot_level_substitutes.has_key(level) and \
               self.pre_pilot_level_substitutes[level]:
            for pre_pilot in self.pre_pilot_level_substitutes[level].keys():
                if member_id in self.pre_pilot_level_substitutes[level][
                    pre_pilot]:
                    res.append(pre_pilot)
        return self.filterExistingUsers(res)

    security.declarePublic('isAssociateSubstituteHere')
    def isAssociateSubstituteHere(self, member_id, level):
        """ Is the given member a substitute at this level ?

        If so returns the member for whom he's been set as a substitute
        """

        res = []

        if self.associate_level_substitutes.has_key(level) and \
               self.associate_level_substitutes[level]:
            for associate in self.associate_level_substitutes[level].keys():
                if member_id in self.associate_level_substitutes[level][
                    associate]:
                    res.append(associate)
        return self.filterExistingUsers(res)

    security.declarePublic('isWatcherSubstituteHere')
    def isWatcherSubstituteHere(self, member_id, level):
        """ Is the given member a substitute at this level ?

        If so returns the member for whom he's been set as a substitute
        """

        res = []

        if self.watcher_level_substitutes.has_key(level) and \
               self.watcher_level_substitutes[level]:
            for watcher in self.watcher_level_substitutes[level].keys():
                if member_id in self.watcher_level_substitutes[level][
                    watcher]:
                    res.append(watcher)
        return self.filterExistingUsers(res)

    ##############################################################
    ##############################################################

    security.declareProtected(ModifyPortalContent,  'setSubstitutesFor')
    def setSubstitutesFor(self, id, level, role, context, with_local_roles=1):
        """Set substitutes in the workflow for a given member/group
        at a given level

        Used at intialization time
        """

        if (role == 'pilot' and self.isPilotSubstituteHere(id, level)
            or role == 'pre_pilot'
            and self.isPrePilotSubstituteHere(id, level)
            or role == 'associate'
            and self.isAssociateSubstituteHere(id, level)
            or role == 'watcher' and self.isWatcherSubstituteHere(id, level)):
            return

        self._p_changed = 1

        all_substitutes = [x for x in self.getSubstitutesFor(id) if x]
        substitutes = []

        for substitute in all_substitutes:
            if substitute.split(':')[0] == 'members':
                substitutes.append(substitute.split(':')[1])
            # XXX check that strange behavior
            elif substitute.split(':')[0] == 'groups' or \
                 substitute.split(':')[0] == 'group':
                substitutes.append('group:'+substitute.split(':')[1])

        for substitute in substitutes:
            if role == 'pilot':
                if substitute not in self.getPilots(level=level):
                    self.pilots[level] += [substitute]
                    LOG("MisillWorkflowEngine : setSubstitutesFor() : PILOTS",
                        DEBUG,
                        self.pilots)
                    # Substitutes struct for pilots
                    if self.pilot_level_substitutes.has_key(level):
                        if self.pilot_level_substitutes[level].has_key(id):
                            if (substitute not in
                                self.pilot_level_substitutes[level][id]):
                                self.pilot_level_substitutes[
                                    level][id] += [substitute]
                        else:
                            self.pilot_level_substitutes[
                                level][id] = [substitute]
                    else:
                        self.pilot_level_substitutes[level] = {}
                        self.pilot_level_substitutes[level][id] = [substitute]
                    if with_local_roles:
                        if not ':' in substitute:
                            context.manage_setLocalRoles(
                                substitute,
                                ['WorkspaceManager'])
                            LOG("MisillWorkflowEngine : Notify Pilots",
                                INFO,
                                substitute)
                            self.notify_mailfolder_event(context,
                                                         'assign_pilot',
                                                         [substitute])
                        else:
                            context.manage_setLocalGroupRoles(
                                substitute,
                                ['WorkspaceManager'])
                            all = self._getGroupMembers(
                                substitute.split(':')[1])
                            LOG("MisillMailManager : Notify Pilots",
                                INFO,
                                all)
                            self.notify_mailfolder_event(context,
                                                         'assign_pilot',
                                                         all)

            elif role == 'pre_pilot':
                if substitute not in self.getPrePilots(level=level):
                    self.pre_pilots[level] += [substitute]
                    # Substitutes struct for pre_pilots
                    if self.pre_pilot_level_substitutes.has_key(level):
                        if self.pre_pilot_level_substitutes[level].has_key(id):
                            if (substitute not in
                                self.pre_pilot_level_substitutes[level][id]):
                                self.pre_pilot_level_substitutes[
                                    level][id] +=[substitute]
                        else:
                            self.pre_pilot_level_substitutes[
                                level][id] = [substitute]
                    else:
                        self.pre_pilot_level_substitutes[level] = {}
                        self.pre_pilot_level_substitutes[
                            level][id] = [substitute]

            elif role =='associate':
                if substitute not in self.getAssociates():
                    self.associates[level] += [substitute]
                    # Substitutes struct for associates
                    if self.associate_level_substitutes.has_key(level):
                        if self.associate_level_substitutes[level].has_key(id):
                            if (substitute not in
                                self.associate_level_substitutes[level][id]):
                                self.associate_level_substitutes[
                                    level][id] +=[substitute]
                        else:
                            self.associate_level_substitutes[
                                level][id] = [substitute]
                    else:
                        self.associate_level_substitutes[level] = {}
                        self.associate_level_substitutes[
                            level][id] = [substitute]
                    if not ':' in substitute:
                        context.manage_setLocalRoles(
                            substitute,
                            ['WorkspaceMember'])
                    else:
                        context.manage_setLocalGroupRoles(
                            substitute,
                            ['WorkspaceMember'])

            elif role == 'watcher':
                if substitute not in self.getWatchers():
                    self.watchers[level] += [substitute]
                    # Substitutes struct for watchers
                    if self.watcher_level_substitutes.has_key(level):
                        if self.watcher_level_substitutes[level].has_key(id):
                            if (substitute not in
                                self.watcher_level_substitutes[level][id]):
                                self.watcher_level_substitutes[
                                    level][id] +=[substitute]
                        else:
                            self.watcher_level_substitutes[
                                level][id] = [substitute]
                    else:
                        self.watcher_level_substitutes[level] = {}
                        self.watcher_level_substitutes[
                            level][id] = [substitute]
                    if not ':' in substitute:
                        context.manage_setLocalRoles(
                            substitute,
                            ['WorkspaceReader'])
                    else:
                        context.manage_setLocalGroupRoles(
                            substitute,
                            ['WorkspaceReader'])

            else:
                pass

        LOG("setSubstitutesFor() reindexObjectSecuroty on",
            DEBUG,
            context.absolute_url())

        context.reindexObjectSecurity()

    security.declareProtected(ModifyPortalContent,  'removeSubstitutesFor')
    def removeSubstitutesFor(self, id, level, role, context):
        """Remove substitutes in the workflow for a given member/group
        at a given level
        """

        self._p_changed = 1

        all_substitutes = [x for x in self.getSubstitutesFor(id) if x]
        substitutes = []

        for substitute in all_substitutes:
            if substitute.split(':')[0] == 'members':
                substitutes.append(substitute.split(':')[1])
            elif substitute.split(':')[0] == 'groups' or \
                 substitute.split(':')[0] == 'group':
                substitutes.append('group:'+substitute.split(':')[1])

        for substitute in substitutes:
            if role == 'pilot':
                if substitute in self.getPilots(level=level):
                    try:
                        self.pilots[level].remove(substitute)
                        # Substitutes struct for pilots
                        self.pilot_level_substitutes[level][id].remove(
                            substitute)
                    except (KeyError, ValueError):
                        pass
                    if not ':' in substitute:
                        context.manage_delLocalRoles(
                            substitute,
                            ['WorkspaceManager'])
                    else:
                        context.manage_delLocalGroupRoles(
                            substitute,
                            ['WorkspaceManager'])

            if role == 'pre_pilot':
                if substitute in self.getPrePilots(level=level):
                    try:
                        self.pre_pilots[level].remove(substitute)
                        # Substitutes struct for pre_pilots
                        self.pre_pilot_level_substitutes[level][id].remove(
                            substitute)
                    except (KeyError, ValueError):
                        pass

            elif role =='associate':
                if substitute in self.getAssociates():
                    try:
                        self.associates[level].remove(substitute)
                        # Substitutes struct for associates
                        self.associate_level_substitutes[level][id].remove(
                            substitute)
                    except (KeyError, ValueError):
                        pass
                    if not ':' in substitute:
                        context.manage_delLocalRoles(
                            substitute,
                            ['WorkspaceMember'])
                    else:
                        context.manage_delLocalGroupRoles(
                            substitute,
                            ['WorkspaceMember'])

            elif role == 'watcher':
                if substitute in self.getWatchers():
                    try:
                        self.watchers[level].remove(substitute)
                        # Substitutes struct for watchers
                        self.watcher_level_substitutes[level][id].remove(
                            substitute)
                    except (KeyError, ValueError):
                        pass
                    if not ':' in substitute:
                        context.manage_delLocalRoles(
                            substitute,
                            ['WorkspaceReader'])
                    else:
                        context.manage_delLocalGroupRoles(
                            substitute,
                            ['WorkspaceReader'])
            else:
                pass
        LOG("removeSubstitutesFor() reindexObjectSecuroty on",
            DEBUG, context.absolute_url())
        context.reindexObjectSecurity()

    def filterExistingUsers(self, list):
        dirs = getToolByName(self, 'portal_directories')
        memberdir = dirs.members
        groupdir = dirs.groups
        res = []
        for each in list:
            if ':' in each:
                name = each.split(':')[1]
                if groupdir.hasEntry(name):
                    res.append(each)
            else:
                if memberdir.hasEntry(each):
                    res.append(each)
        return res            
    
InitializeClass(Substitutes)
