# -*- coding: ISO-8859-15 -*-
# Copyright (c) 2004 Nuxeo SARL <http://nuxeo.com>
# Copyright (c) 2004 CGEY <http://cgey.com>
# Copyright (c) 2004 Ministre de L'intrieur (MISILL)
#               <http://www.interieur.gouv.fr/>
# Author: Julien Anguenot <ja@nuxeo.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id: WorkflowManager.py,v 1.18 2004/07/01 17:39:53 regebro Exp $

__author__ = "Julien Anguenot <ja@nuxeo.com>"

"""WorkflowManager class.

Core workflow class implementing the dynamic workflow implementation.  It
provides transition guards, and actions to be taken at transition execution
time.

Check the workflows provides within Workflows sub-folder to see how it is put
into action
"""

from types import StringType

from Globals import InitializeClass
import ExtensionClass
from AccessControl import ClassSecurityInfo
from Acquisition import aq_base

from Products.CMFCore.WorkflowCore import WorkflowException
from Products.CMFCore.CMFCorePermissions import ModifyPortalContent, View
from Products.CMFCore.utils import getToolByName

from Substitutes import Substitutes

from zLOG import LOG, DEBUG, INFO

#
# This code needs a refactoring for several reason :
# - Moving some of the logic within the CPSCore to avoid the local roles
#    management.
#  - See how to reduce the complexity since the MISILL implementation was quite
#    complex
#  - And because it's needed too :)
#
#  JA (05/07/2004)
#

class WorkflowManager(Substitutes):
    """WorkflowManager class.

    Core workflow class implementing the dynamic workflow implementation.  It
    provides transition guards, and actions to be taken at transition execution
    time.

    Check the workflows provides within Workflows sub-folder to see how it is
    put into action
    """

    security = ClassSecurityInfo()

    def __init__(self, **kw):
        """Constructor

        Init of the assignment lists.
        Will be checked through class methods with
        the guards. We need it cause we can't only manage
        that aspect with workflow based permissions.
        """

        # Keep track at which level we are.
        self.current_level = 0

        # Pre pilots
        self.pre_pilots = {0: []}

        # Pilot List
        creator_id = kw['current_member_id']
        self.pilots = {0: [creator_id]}

        # Associate List
        self.associates = {0: []}

        # Watcher List
        self.watchers = {0: []}

        # States that don't need pre-assignment
        self.no_pre_assignment = ['writing', 'validating']

        Substitutes.__init__(self)

    ######################################################
    ######################################################

    def _getWorkflowId(self):
        """Returns the workflow the class is driving
        """
        return self.workflow_id


    def _getIndex(self, list=[], item=None):
        """Return the index of an item in a list.

        None if no index.
        """
        if item not in list:
            return None
        i = 0
        for j in list:
            if j == item:
                return i
            i +=1
        return i

    def _getFormerState(self, context):
        """Return the former state.

        Used to come back from the closing state to the former one.
        """
        wftool = context.portal_workflow
        review_history = wftool.getInfoFor(context, 'review_history')
        review_history.reverse()

        # Check if this
        if len(review_history) > 1:
            return review_history[1]['review_state']
        else:
            return None

    def _log(self):
        """DEBUG
        """
        return
        LOG("CURRENT LEVEL",
            DEBUG,
            self.getCurrentLevel())
        LOG("PRE PILOTS",
            DEBUG,
            self.getPrePilots(level="all"))
        LOG("PRE PILOTS MEMBER IDS",
            DEBUG,
            self.getPrePilotMemberIds(level="all"))
        LOG("PRE PILOTS SUBSTITUTES IDS",
            DEBUG,
            self.pre_pilot_level_substitutes)
        LOG("PILOTS",
            DEBUG,
            self.getPilots(level="all"))
        LOG("PILOTS MEMBER IDS",
            DEBUG,
            self.getPilotMemberIds(level="all"))
        LOG("PILOTS SUBSTITUTES IDS ",
            DEBUG,
            self.pilot_level_substitutes)
        LOG("ASSOCIATES",
            DEBUG,
            self.getAssociates())
        LOG("WATCHERS",
            DEBUG,
            self.getWatchers())


    ######################################################
    ######################################################

    def _insertPrePilots(self, left=0, right=0, id=None, context=None):
        """Insert a pre pilot at a given position within the
        assignment tree.
        """

        self._p_changed = 1

        # Only used for the substitute function.
        # Not within the algo
        level = 0

        if left == right:
            if self.pre_pilots.has_key(left):
                if id not in self.pre_pilots[left]:
                    self.pre_pilots[left] += [id]
            else:
                self.pre_pilots[left] = [id]
            level = left
            if left == self.getCurrentLevel():
                #
                # If at current level we give rights straight
                #
                self._assignLocalRoles(id, 'pilote', context)
        else:
            keys = self.pre_pilots.keys()
            if left < self.getCurrentLevel():
                keys = [x for x in keys if x <= left]
                keys.sort()
                for key in keys:
                    self.pre_pilots[key-1] = self.pre_pilots[key]
                    for pre_pilot in self.pre_pilots[key-1]:
                        self.removeSubstitutesFor(pre_pilot, key,
                                                  'pre_pilot', context)
                        self.setSubstitutesFor(pre_pilot, key-1,
                                               'pre_pilot', context)
                self.pre_pilots[left] = [id]
                level = left
            elif left >= self.getCurrentLevel():
                keys = [x for x in keys if x > left]
                keys.sort()
                keys.reverse()
                for key in keys:
                    self.pre_pilots[key+1] = self.pre_pilots[key]
                    for pre_pilot in self.pre_pilots[key+1]:
                        self.removeSubstitutesFor(pre_pilot, key,
                                                  'pre_pilot', context)
                        self.setSubstitutesFor(pre_pilot, key+1,
                                               'pre_pilot', context)
                self.pre_pilots[left+1] = [id]
                level = left+1

        # Substitutes for the one we just add in the WF
        self.setSubstitutesFor(id, level, 'pre_pilot', context)

    def _insertPilots(self, left=0, right=0, id=None, context=None):
        """Insert a pilot at a given position within the
        assignment tree.
        """

        self._p_changed = 1

        # Only used for the substitute function.
        # Not within the algo
        level = 0

        if left == right:
            if self.pilots.has_key(left):
                if id not in self.pilots[left]:
                    self.pilots[left] += [id]
            else:
                self.pilots[left] = [id]
            level = left
        else:
            keys = self.pilots.keys()
            if left < self.getCurrentLevel():
                keys = [x for x in keys if x <= left]
                keys.sort()
                for key in keys:
                    self.pilots[key-1] = self.pilots[key]
                    for pilot in self.pilots[key-1]:
                        self.removeSubstitutesFor(pilot, key-1,
                                                  'pilot', context)
                        self.setSubstitutesFor(pilot, key-1,
                                               'pilot', context)
                self.pilots[left] = [id]
                level = left
            elif left >= self.getCurrentLevel():
                keys = [x for x in keys if x > left]
                keys.sort()
                keys.reverse()
                for key in keys:
                    self.pilots[key+1] = self.pilots[key]
                    for pilot in self.pilots[key+1]:
                        self.removeSubstitutesFor(pilot, key,
                                                  'pilot', context)
                        self.setSubstitutesFor(pilot, key+1,
                                               'pilot', context)
                self.pilots[left+1] = [id]
                level = left+1
        # Local Roles
        self._assignLocalRoles(id, 'pilote', context)
        # Substitutes for the one we just add in the WF
        self.setSubstitutesFor(id, level, 'pilot', context)

    def _assignLocalRoles(self, member_id, role, context):
        """Assign local roles to the given member_id.

        Not on self cause we are in the repository but on
        the proxy.
        """

        if role == "associate":
            if not ':' in member_id :
                context.manage_setLocalRoles(member_id,
                                             ['WorkspaceMember'])
            else:
                context.manage_setLocalGroupRoles(member_id.split(':')[1],
                                                  ['WorkspaceMember'])
        elif role == "watcher":
            if not ':' in member_id:
                context.manage_setLocalRoles(member_id,
                                             ['WorkspaceReader'])
            else:
                context.manage_setLocalGroupRoles(member_id.split(':')[1],
                                                  ['WorkspaceReader'])
        elif role == "pilote":
            if not ':' in member_id:
                context.manage_setLocalRoles(member_id,
                                             ['WorkspaceManager'])
                LOG("MisillWorkflowEngine : Notify Pilots",
                    INFO,
                    member_id)
                self.notify_mailfolder_event(context,
                                             'assign_pilot',
                                             [member_id])
            else:
                context.manage_setLocalGroupRoles(member_id.split(':')[1],
                                                  ['WorkspaceManager'])
                all = self._getGroupMembers(member_id.split(':')[1])
                LOG("MisillWorkflowEngine : Notify Pilots",
                    INFO,
                    all)
                self.notify_mailfolder_event(context,
                                             'assign_pilot',
                                             all)
        else:
            pass

        context.reindexObjectSecurity()

    def _removeLocalRoles(self, member_id, context):
        """Removing local roles to a givne member in a context
        """

        self._p_changed = 1

        if not ':' in member_id:
            context.manage_delLocalRoles([member_id])
        else:
            context.manage_delLocalGroupRoles([member_id.split(':')[1]])

        context.reindexObjectSecurity()

    def _updateLocalRoles(self, context, order='inc'):
        """Updating local roles while validating an aassignement
        """

        self._p_changed = 1

        current_level = self.getCurrentLevel()

        if order == 'inc':
            for member in self.pre_pilots[current_level+1]:
                self._assignLocalRoles(member,
                                       'pilote',
                                       context)
        else:
            for member in self.pre_pilots[current_level-1]:
                self._assignLocalRoles(member,
                                       'pilote',
                                       context)
        if context is not None:
            context.reindexObjectSecurity()

    #######################################################
    #######################################################

    security.declarePublic("getCurrentLevel")
    def getCurrentLevel(self):
        """Returns the current workflow level
        """
        return self.current_level

    #######################################################
    #######################################################

    security.declarePublic("getPilots")
    def getPilots(self, level=None):
        """Get the pilots

        Order matters
        """

        if level is None:
            if self.pilots.has_key(self.getCurrentLevel()):
                return self.filterExistingUsers(self.pilots[self.getCurrentLevel()])
            else:
                return []
        elif level == "all":
            res = {}
            for key, value in self.pilots.items():
                res[key] = self.filterExistingUsers(value)
            return res
        else:
            if self.pilots.has_key(level):
                return self.filterExistingUsers(self.pilots[level])
            else:
                return []

    security.declarePublic('getPilotMemberIds')
    def getPilotMemberIds(self, level=None):
        """ Get the pilot member ids

        We look inside the group members of they are
        """
        pilotes = self.getPilots(level)
        pilote_members = []

        if level != 'all':
            for pilot in pilotes:
                if not ':' in pilot:
                    pilote_members.append(pilot)
                else:
                    for group_pilot in self._getGroupMembers(
                        pilot.split(':')[1]):
                        pilote_members.append(group_pilot)
        else:
            for key in pilotes.keys():
                for pilot in pilotes[key]:
                    if not ':' in pilot:
                        pilote_members.append(pilot)
                    else:
                        for group_pilot in self._getGroupMembers(
                            pilot.split(':')[1]):
                            pilote_members.append(group_pilot)

        return pilote_members


    security.declarePublic('getPilotsSubstitutesExcluded')
    def getPilotsSubstitutesExcluded(self, level=None):
        """ Get the *real* pilots. No substitutes
        """

        if level is None or level == 'all':
            all_pilots = {}
            for key in self.getPilots(level=level).keys():
                for member_id in self.getPilots(level=level)[key]:
                    if not self.isPilotSubstituteHere(member_id, key):
                        pt = all_pilots.get(key, [])
                        pt += [member_id]
                        all_pilots[key] = pt
        else:
            all_pilots = []
            for member_id in self.getPilots(level=level):
                if not self.isPilotSubstituteHere(member_id, level):
                    all_pilots.append(member_id)
        return all_pilots

    ####################################################
    ####################################################

    security.declarePublic("getPrePilots")
    def getPrePilots(self, level=None):
        """Get the Pre-Pilots.

        Order matters
        """

        if level is None:
            if self.pre_pilots.has_key(self.current_level):
                return self.filterExistingUsers(self.pre_pilots[self.current_level])
            else:
                return []
        elif level == "all":
            res = {}
            for key, value in self.pre_pilots.items():
                res[key] = self.filterExistingUsers(value)
            return res
        else:
            if self.pre_pilots.has_key(level):
                return self.filterExistingUsers(self.pre_pilots[level])
            else:
                return []

    security.declarePublic('getPrePilotMemberIds')
    def getPrePilotMemberIds(self, level=None):
        """ Get the pre pilot member ids

        We look inside the group members of they are
        """
        pre_pilots = self.getPrePilots(level)
        pre_pilot_members = []

        if level != 'all':
            for pilot in pre_pilots:
                if not ':' in pilot:
                    pre_pilot_members.append(pilot)
                else:
                    for group_pilot in self._getGroupMembers(
                        pilot.split(':')[1]):
                        pre_pilot_members.append(group_pilot)
        else:
            for key in pre_pilots.keys():
                for pilot in pre_pilots[key]:
                    if not ':' in pilot:
                        pre_pilot_members.append(pilot)
                    else:
                        for group_pilot in self._getGroupMembers(
                            pilot.split(':')[1]):
                            pre_pilot_members.append(group_pilot)

        return pre_pilot_members

    security.declarePublic('getPrePilotsSubstitutesExcluded')
    def getPrePilotsSubstitutesExcluded(self, level=None):
        """ Get the *real* pre pilots. No substitutes
        """

        if level is None or level == 'all':
            all_pre_pilots = {}
            for key in self.getPrePilots(level=level).keys():
                for member_id in self.getPrePilots(level=level)[key]:
                    if not self.isPrePilotSubstituteHere(member_id, key):
                        pt = all_pre_pilots.get(key, [])
                        pt += [member_id]
                        all_pre_pilots[key] = pt
        else:
            all_pre_pilots = []
            for member_id in self.getPilots(level=level):
                if not self.isPilotSubstituteHere(member_id, level):
                    all_pre_pilots.append(member_id)
        return all_pre_pilots

    ######################################################
    ######################################################

    security.declarePublic("getAssociates")
    def getAssociates(self):
        """Get the associates

        Order doesn't matter. Only at level 0 right now.
        No associate hierarchies
        """

        level = 0
        return self.filterExistingUsers(self.associates[level])

    security.declarePublic("getAssociateMemberIds")
    def getAssociateMemberIds(self):
        """Get all the associate member ids

        Order doesn't matter. Only at level 0 right now.
        No associate hierarchies
        """

        level = 0
        res = []

        for associate in self.associates[level]:
            if not ':' in associate:
                res.append(associate)
            else:
                for group_associate in self._getGroupMembers(
                    associate.split(':')[1]):
                    res.append(group_associate)
        return self.filterExistingUsers(res)

    security.declarePublic('getAssociatesSubstitutesExcluded')
    def getAssociatesSubstitutesExcluded(self, level=None):
        """ Get the *real* associates. No substitutes
        """
        all_associates = []
        if level is None or level == 'all':
            for key in self.getAssociates(level=level).keys():
                for member_id in self.getAssociates(level=level)[key]:
                    if not self.isAssociateSubstituteHere(member_id, key):
                        all_associates.append(member_id)
        return all_associates

    #####################################################
    #####################################################

    security.declarePublic("getWatchers")
    def getWatchers(self):
        """Get the watchers

        Order doesn't matter. Only at level 0 right now.
        No watcher hierarchies
        """

        level = 0
        return self.filterExistingUsers(self.watchers[level])

    security.declarePublic("getWatcherMemberIds")
    def getWatcherMemberIds(self):
        """Get all the watcher member ids

        Order doesn't matter. Only at level 0 right now.
        No associate hierarchies
        """

        level = 0
        res = []

        for watcher in self.watchers[level]:
            if not ':' in watcher:
                res.append(watcher)
            else:
                for group_watcher in self._getGroupMembers(
                    watcher.split(':')[1]):
                    res.append(group_watcher)
        return self.filterExistingUsers(res)

    security.declarePublic('getWatchersSubstitutesExcluded')
    def getWatchersSubstitutesExcluded(self, level=None):
        """ Get the *real* watchers. No substitutes
        """
        all_watchers = []
        if level is None or level == 'all':
            for key in self.getWatchers(level=level).keys():
                for member_id in self.getWatchers(level=level)[key]:
                    if not self.isWatchersubstituteHere(member_id, key):
                        all_watchers.append(member_id)
        return all_watchers

    #############################################################
    #############################################################

    security.declarePublic("getAllPreAssigned")
    def getAllPreAssigned(self, level=None):
        """Returns pre assigned member structs

        Including Assocites/Watchers (Conceptual)
        """

        struct = []

        pt_level = self.current_level
        if level is not None:
            pt_level = level
            for i in self.getPrePilots(level=pt_level):
                struct.append({'login':i,
                               'role':'pilote',
                               'level':pt_level})
        else:
            for key in self.pre_pilots.keys():
                for i in self.getPrePilots(level=key):
                    struct.append({'login':i,
                                   'role':'pilote',
                                   'level':key})

        for member in self.getAssociates():
            struct.append({'login':member,
                           'role':'associate',
                           'level':0})

        for member in self.getWatchers():
            struct.append({'login':member,
                           'role':'watcher',
                           'level':0})

        return struct

    security.declarePublic("getAllPreAssignedMemberIds")
    def getAllPreAssignedMemberIds(self, level=None):
        """Returns pre assigned member id structs

        Including Assocites/Watchers (Conceptual)
        """

        struct = []

        pt_level = self.current_level
        if level is not None:
            pt_level = level
            for i in self.getPrePilotMemberIds(level=pt_level):
                struct.append({'login':i,
                               'role':'pilote',
                               'level':pt_level})
        else:
            for key in self.pre_pilots.keys():
                for i in self.getPrePilotMemberIds(level=key):
                    struct.append({'login':i,
                                   'role':'pilote',
                                   'level':key})

        for member in self.getAssociateMembreIds():
            struct.append({'login':member,
                           'role':'associate',
                           'level':0})

        for member in self.getWatcherMembreIds():
            struct.append({'login':member,
                           'role':'watcher',
                           'level':0})

        return struct

    ###########################################################
    ###########################################################

    security.declarePublic("getAllAssigned")
    def getAllAssigned(self, level=None):
        """Return assigned member structs

        Including Associates/Watchers (Conceptual)
        """

        self._log()

        struct = []

        pt_level = self.current_level
        if level is not None:
            pt_level = level
            for i in self.getPilots(level=pt_level):
                struct.append({'login':i,
                               'role':'pilote',
                               'level':pt_level})
        else:
            for key in self.getPilots(level='all').keys():
                for i in self.getPilots(level=key):
                    struct.append({'login':i,
                                   'role':'pilote',
                                   'level':key})

        for member in self.getAssociates():
            struct.append({'login':member,
                           'role':'associate',
                           'level':0})

        for member in self.getWatchers():
            struct.append({'login':member,
                           'role':'watcher',
                           'level':0})

        return struct

    security.declarePublic("getAllAssignedMemberIds")
    def getAllAssignedMemberIds(self, level=None):
        """Return assigned member ids structs

        Including Associates/Watchers (Conceptual)
        """

        struct = []

        pt_level = self.current_level
        if level is not None:
            pt_level = level
            for i in self.getPilotMemberIds(level=pt_level):
                struct.append({'login':i,
                               'role':'pilote',
                               'level':pt_level})
        else:
            for key in self.getPilots(level='all').keys():
                for i in self.getPilotMemberIds(level=key):
                    struct.append({'login':i,
                                   'role':'pilote',
                                   'level':key})

        for member in self.getAssociateMemberIds():
            struct.append({'login':member,
                           'role':'associate',
                           'level':0})

        for member in self.getWatcherMemberIds():
            struct.append({'login':member,
                           'role':'watcher',
                           'level':0})

        return struct

    ##################################################################
    ##################################################################

    security.declareProtected(ModifyPortalContent, "deletePreAssigned")
    def deletePreAssigned(self, www={} , context=None, level=None):
        """Delete given pre-assigned given their ids

        In this case, no need to remove local roles since they don't have any
        at that time.
        """

        self._p_changed = 1

        LOG("DELETE PRE ASSIGNED", DEBUG, www)
        self._log()

        pt_level = self.current_level

        if level is not None:
            pt_level = level

        if self.pre_pilots.has_key(pt_level):

            #
            # Removing the given user a a given level
            # As well, removing substitutes.
            #

            pre_pilots = self.pre_pilots[pt_level]
            for i in www:
                index = self._getIndex(pre_pilots, i)
                if index is not None:
                    # Substitutes
                    self.removeSubstitutesFor(i, pt_level,
                                              'pre_pilot', context)
                    index = self._getIndex(pre_pilots, i)
                    del self.pre_pilots[pt_level][index]
                    pre_pilots = self.pre_pilots[pt_level]
            self.pre_pilots[pt_level] = pre_pilots

            #
            # Filling the gab if needed
            # In the case of the user removed was the only
            # one at this level
            #

            if (self.pre_pilots[pt_level] == [] and
                self.pilots.get(pt_level, []) == []):
                keys = [x for x in self.pre_pilots.keys() if x >= pt_level]
                keys.sort()
                for key in keys:
                    self.pre_pilots[key] = self.pre_pilots.get(key+1, [])
                    for pre_pilot in self.pre_pilots[key]:
                        self.removeSubstitutesFor(pre_pilot, key+1,
                                                  'pre_pilot', context)
                        self.setSubstitutesFor(pre_pilot, key,
                                               'pre_pilot', context)

    security.declareProtected(ModifyPortalContent, "deleteAssigned")
    def deleteAssigned(self, www={} ,context=None, level=None):
        """Delete the given assignes given their ids

        Removing assignees from the assignees list and then removing the
        corresponding local role in the context.
        """

        self._p_changed = 1

        ## DEBUG
        self._log()

        pt_level = self.current_level

        if level is not None:
            pt_level = level

        if context is not None:

            #
            # Removing the given user a a given level
            # As well, removing substitutes.
            #

            if self.pilots.has_key(pt_level):
                pilots = self.pilots[pt_level]
                for i in www:
                    index = self._getIndex(pilots, i)
                    if (index is not None and
                        self.Creator() != i):
                        # Substitutes
                        self.removeSubstitutesFor(i, pt_level,
                                                  'pilot', context)
                        index = self._getIndex(pilots, i)
                        del self.pilots[pt_level][index]
                        pilots = self.pilots[pt_level]
                self.pilots[pt_level] = pilots

            #
            # Filling the gab if needed
            # In the case of the user removed was the only
            # one at this level
            #

            if self.pilots[pt_level] == []:
                keys = [x for x in self.pilots.keys() if x <= pt_level]
                keys.sort()
                keys.reverse()
                for key in keys:
                    self.pilots[key] = self.pilots.get(key-1, [])
                    for pilot in self.pilots[key]:
                        self.removeSubstitutesFor(pilot, key-1,
                                                  'pilot', context)
                        self.setSubstitutesFor(pilot, key, 'pilot', context)

    security.declareProtected(ModifyPortalContent, "deleteAssociates")
    def deleteAssociates(self, www={}, context=None, level=None):
        """ Deletes given associates give their ids

        Removing associates from the associates list and then removing the
        corresponding local role in the context.
        """

        if context is not None:
            associates = self.getAssociates()
            for i in www:
                index = self._getIndex(associates, i)
                if index is not None:
                    self.removeSubstitutesFor(i, 0, 'associate', context)
                    index = self._getIndex(associates, i)
                    del self.associates[0][index]
                    self._removeLocalRoles(i, context)
                    associates = self.getAssociates()
            self.associates[0] = associates
        self._p_changed = 1

    security.declareProtected(ModifyPortalContent, "deleteWatchers")
    def deleteWatchers(self, www={}, context=None, level=None):
        """ Deletes given watchers given their ids

        Removing associates from the associates list and then removing the
        corresponding local role in the context.
        """

        if context is not None:
            watchers = self.getWatchers()
            for i in www:
                index = self._getIndex(watchers, i)
                if index is not None:
                    self.removeSubstitutesFor(i, 0, 'watcher', context)
                    index = self._getIndex(watchers, i)
                    del self.watchers[0][index]
                    self._removeLocalRoles(i, context)
                    watchers = self.getWatchers()
            self.watchers[0] = watchers
        self._p_changed = 1

    #######################################################################
    #######################################################################

    security.declarePublic("getMergedAssignedTree")
    def getMergedAssignedTree(self):
        """Return the complete pilots tree of assignement.

        It's a merge in between already assigned ones (validated ones)
        and the pre-assigned ones.

        We add a flag to know if they are validated already
        """

        self._log()

        pilots = self.getPilotsSubstitutesExcluded(level="all")
        pre_pilots = self.getPrePilotsSubstitutesExcluded(level="all")

        pilots_tree = {}
        to_validate = {}

        pilote_keys = pilots.keys()
        if pilote_keys == []:
            min_pilote_keys = 0
            max_pilote_keys = 0
        else:
            min_pilote_keys = min(pilote_keys)
            max_pilote_keys = max(pilote_keys)

        pre_pilote_keys = pre_pilots.keys()
        if pre_pilote_keys == []:
            min_pre_pilote_keys = 0
            max_pre_pilote_keys = 0
        else:
            min_pre_pilote_keys = min(pre_pilote_keys)
            max_pre_pilote_keys = max(pre_pilote_keys)

        start = min(min_pilote_keys, min_pre_pilote_keys)
        stop  = max(max_pilote_keys, max_pre_pilote_keys)

        for key in range(start, stop+1):
            current_level = []
            if pilots.has_key(key):
                current_level += pilots[key]
            if pre_pilots.has_key(key):
                current_level += pre_pilots[key]
                to_validate[key] = pre_pilots[key]

            # We don't want any empty list
            if current_level:
                pilots_tree[key] = current_level


        # Removing duplicata if they are
        for key in pilots_tree.keys():
            level = pilots_tree[key]
            new_level = []
            for member in level:
                if member not in new_level:
                    new_level.append(member)
            pilots_tree[key] = new_level

        return [pilots_tree, to_validate]

    ########################################################

    def _getGroupMembers(self, group_id):
        """ Returns the list of members given a group id
        """

        res = []
        dirtool = self.portal_directories
        group = dirtool.groups.getEntry(group_id)
        for member in group['members']:
            if member not in res:
                res.append(member)
        subgroups = group['subgroups']
        if subgroups:
            for subgroup in subgroups:
                tmp = []
                tmp = self._getGroupMembers(subgroup)
                for member in tmp:
                    if member not in res:
                        res.append(member)
        return res

    def _isUserMemberOfGroup(self, member_id, group_id):
        """ Does a given member belong to a given group ?
        """
        return member_id in self._getGroupMembers(group_id)

    ########################################################

    security.declarePublic("getCurrentMemberRoles")
    def getCurrentMemberRoles(self, level=None):
        """Return the member roles in here as Strings.

        Whatever pilots (and levels) / associates / watchers
        """

        self._log()

        portal = self.portal_url.getPortalObject()
        member = portal.portal_membership.getAuthenticatedMember()
        member_id = member.getMemberId()

        res = []

        if level is not None:
            pt_level = level
            for pilot in self.getPilots(level=pt_level):
                dummy = 0
                if not ':' in pilot:
                    if member_id == pilot:
                        dummy = 1
                else:
                    if self._isUserMemberOfGroup(
                        member_id, pilot.split(':')[1]):
                        dummy = 1
                if dummy:
                    res.append('P'+str(pt_level))
        else:
            pt_level = "all"
            for key in self.getPilots(level="all").keys():
                for pilot in self.getPilots(level="all")[key]:
                    dummy = 0
                    if not ':' in pilot:
                        if member_id == pilot:
                            dummy = 1
                    else:
                        if self._isUserMemberOfGroup(
                            member_id, pilot.split(':')[1]):
                            dummy = 1
                    if dummy:
                        res.append('P'+str(key))

        for associate in self.getAssociates():
            dummy = 0
            if not ':' in associate:
                if member_id == associate:
                    dummy = 1
            else:
                if self._isUserMemberOfGroup(
                    member_id, associate.split(':')[1]):
                    dummy = 1
            if dummy:
                res += ['Associate']

        for watcher in self.getWatchers():
            dummy = 0
            if not ':' in watcher:
                if member_id == watcher:
                    dummy = 1
            else:
                if self._isUserMemberOfGroup(
                    member_id, watcher.split(':')[1]):
                    dummy = 1
            if dummy:
                res += ['Watcher']

        return res

    ####################################################################
    ####################################################################

    security.declarePublic("canReject")
    def canReject(self, context=None):
        """Reject transition guard.

        You can reject the assignement except if there's nobody above you.
        """
        if context is not None:
            if (self.portal_membership.getAuthenticatedMember().getMemberId()
                in self.getPilotMemberIds(level=self.getCurrentLevel()) and
                self.getPilots(level=self.getCurrentLevel() - 1)):
                return 1
        return 0

    security.declarePublic("canAssign")
    def canAssign(self, context=None):
        """Validate transition.

        Will be possible to validate only if they're some people
        pre-affected and if you are a pilote or a Manager.
        """
        if context is not None:
            member = context.portal_membership.getAuthenticatedMember()
            member_id = member.getMemberId()
            if (self.getPrePilots(level='all').has_key(
                self.getCurrentLevel()+1) and
                self.getPrePilots(level='all')[self.getCurrentLevel()+1] and
                (member_id in
                 self.getPilotMemberIds(level=self.getCurrentLevel()) or
                 member.has_role(['Manager',], context))):
                return 1
        return 0

    security.declarePublic("canInitializeReply")
    def canInitializeReply(self, context=None):
        """Initialize_reply transition guard

        May be realized if you're a pilot or manager when the workflow is in an
        assigning state.
        """

        if context is not None:
            member = context.portal_membership.getAuthenticatedMember()
            if member.has_role(['WorkspaceManager', 'Manager',], context):
                return 1
        return 0

    security.declarePublic("canValidate")
    def canValidate(self, context=None):
        """Validate transition guard

        May be possible to do that if you're a pilot or manager
        in a validating state and if you're not the latest having to
        validate.
        """

        if context is not None:
            member = context.portal_membership.getAuthenticatedMember()
            member_id = member.getMemberId()
            if (member.has_role(['WorkspaceManager', 'Manager'], context) and
                (member_id in
                 self.getPilotMemberIds(level=self.getCurrentLevel()) or
                 member.has_role(['Manager',], context)) and
                (self.pilots.has_key(self.getCurrentLevel()-1) and
                 self.getPilots(level=self.getCurrentLevel()-1) != [])):
                return 1
        return 0

    security.declarePublic("canRejectValidate")
    def canRejectValidate(self):
        """Reject_validate transition guard
        """

        mtool = self.portal_membership
        member_id = mtool.getAuthenticatedMember().getMemberId()

        return ((member_id in
                 self.getPilotMemberIds(level=self.getCurrentLevel()) or
                 member_id in
                 self.getPrePilotMemberIds(level=self.getCurrentLevel())) and
                (self.getPrePilots(level=self.getCurrentLevel()+1) or
                 self.getPilots(level=self.getCurrentLevel()+1)))

    security.declarePublic("canReset")
    def canReset(self, context=None):
        """Reset transition guard

        Only the creator of the mail folder or the manager are allowed to do
        that
        """

        if context is not None:
            member = self.portal_membership.getAuthenticatedMember()
            member_id = member.getMemberId()
            return member_id == self.Creator() or \
                   member.has_role("Manager", context)
        return 0

    security.declarePublic("canClose")
    def canClose(self, context=None):
        """Close transition

        All pilots can do it.
        """

        if context is not None:
            member = self.portal_membership.getAuthenticatedMember()
            return member.has_role(['WorkspaceManager', 'Manager',], context)
        return 0

    security.declarePublic("canUnClose2waiting")
    def canUnClose2waiting(self, context=None):
        """unclose2waiting transition guard

        From a close state to the waiting former state
        """

        if context is not None:
            return self._getFormerState(context) == "waiting"
        return 0

    security.declarePublic("canUnClose2assigning")
    def canUnClose2assigning(self, context=None):
        """unclose2assigning transition guard

        From a close state to the assigning former state
        """

        if context is not None:
            return self._getFormerState(context) == "assigning"
        return 0

    security.declarePublic("canUnClose2writing")
    def canUnClose2writing(self, context=None):
        """unclose2writing transition guard

        From a close state to the writing former state
        """

        if context is not None:
            return self._getFormerState(context) == "writing"
        return 0

    security.declarePublic("canUnCLose2validating")
    def canUnClose2validating(self, context=None):
        """unclose2validating transition guard

        For the close state to the validating former state
        """

        if context is not None:
            return self._getFormerState(context) == "validating"
        return 0

    security.declarePublic("canPublish")
    def canPublish(self, context=None):
        """copy_sumbit transition

        It's possible to publish when the mailfolder is closed
        for pilot.
        """

        if context is not None:
            if ((getattr(context, context.getOutGoingMailTypeId(), 0) and
                 self.portal_membership.getAuthenticatedMember().getMemberId()
                 in self.getPilotMemberIds(level=self.getCurrentLevel()) and
                 not self.getPilots(level=self.getCurrentLevel() - 1))):
                return 1
        return 0

    #######################################################################
    #######################################################################

    security.declareProtected(ModifyPortalContent, "doReinit")
    def doReinit(self, context=None, comment=''):
        """Reset transition

        Re-init the workflow and all the parameters of the mailfolder.  We keep
        the history, comments and all the documents inside the mailfolder.
        """

        self._p_changed = 1

        if context is not None:
            wftool = context.portal_workflow
            try:
                wftool.doActionFor(context,
                                   "reset",
                                   wf_id=self._getWorkflowId(),
                                   comment='')
            except (AttributeError, WorkflowException):
                pass

            #
            # Flush *all* the assigned list
            #

            for key in self.pre_pilots.keys():
                for pre_pilot in self.pre_pilots[key]:
                    self._removeLocalRoles(pre_pilot, context)
                    self.removeSubstitutesFor(pre_pilot, 0,
                                              'pre_pilot', context)
            self.pre_pilots = {0:[]}

            for associate in self.associates[0]:
                self._removeLocalRoles(associate, context)
                self.removeSubstitutesFor(associate, 0,
                                          'associate', context)
            self.associates = {0:[]}

            for watcher in self.watchers[0]:
                self._removeLocalRoles(watcher, context)
                self.removeSubstitutesFor(watcher, 0,
                                          'watcher', context)
            self.watchers = {0:[]}

            #
            # Flush pilots
            #

            for key in self.pilots.keys():
                for pilot in self.pilots[key]:
                    self._removeLocalRoles(pilot, context)
                    self.removeSubstitutesFor(pilot, 0, 'pilot', context)

            #
            # Flush *all* the substitutes lists
            # Notice, they're suppose to be empty now.
            #

            self.pilot_level_substitutes = {}
            self.pre_pilots_level_susbtitutes = {}
            self.associate_level_subsitutes = {}
            self.watcher_level_susbtitutes = {}

            #
            # Creator takes the power back
            #

            self.pilots = {0:[self.Creator()]}
            self.grantOwnerManagerRole(context)
            self.setSubstitutesFor(self.Creator(), 0, 'pilot', context)
            self.current_level = 0

    security.declareProtected(ModifyPortalContent, "doPreAssign")
    def doPreAssign(self, www={}, context=None):
        """pre_assign  transition

        XXX dirty method -> refactoring
        """

        self._p_changed = 1

        if context is not None:
            try:
                context.portal_workflow.doActionFor(
                    context,
                    'pre_assign',
                    wf_id=self._getWorkflowId(),)
            except (AttributeError, WorkflowException,):
                pass
            for item in www:
                level = item['level']
                if isinstance(level, StringType):
                    split = level.split(':')
                    left = int(split[0])
                    right = int(split[1])
                else:
                    left = level
                    right = level
                if item['role'] == 'pilote':
                    wftool = context.portal_workflow
                    review_state = wftool.getInfoFor(
                        context,
                        'review_state',
                        wf_id=self._getWorkflowId())

                    #
                    # Here adding a pre-pilot except if we are following a
                    # transition that doesn't need any.  As well, when addding
                    # people at the current level or above they are straight
                    # assigned as pilots
                    #

                    if (left > self.getCurrentLevel() or
                        right > self.getCurrentLevel()):
                        LOG("MisillWorkflowEngine : inserPrePitlots()",
                            DEBUG,
                            item['login'])
                        self._insertPrePilots(left=left,
                                              right=right,
                                              id=item['login'],
                                              context=context)
                    else:
                        LOG("MisillWorkflowEngine : inserPilote()",
                            DEBUG,
                            item['login'])
                        context.doAddPilot(item['login'])
                        self._insertPilots(left=left,
                                           right=right,
                                           id=item['login'],
                                           context=context)
                    #
                    # Is a pilote now so flushing the previous roles if he had.
                    #

                    index = None
                    index = self._getIndex(self.getAssociates(), item['login'])
                    if index is not None:
                        del self.associates[0][index]
                        self._removeLocalRoles(item['login'], context)
                    index = None
                    index = self._getIndex(self.getWatchers(), item['login'])
                    if index is not None:
                        del self.watchers[0][index]
                        self._removeLocalRoles(item['login'], context)

                #
                # No need to give him an associate role if he's already a
                # pilote
                #

                elif item['role'] == 'associate':
                    if item['login'] not in self.getAssociates() + \
                           [x['login'] for x in self.getAllAssigned()] + \
                           [x['login'] for x in self.getAllPreAssigned()]:
                        self.associates[0] += [item['login']]
                        self._assignLocalRoles(item['login'],
                                               'associate',
                                               context)
                        self.setSubstitutesFor(item['login'],
                                               0,
                                               'associate',
                                               context,
                                               with_local_roles=1)
                    index = None
                    index = self._getIndex(self.getWatchers(), item['login'])
                    if index is not None:
                        del self.watchers[0][index]
                        self._removeLocalRoles(item['login'], context)

                #
                # No need to give him a watcher role if he's already
                # a pilote or associate role.
                #

                elif item['role'] == 'watcher':
                    if item['login'] not in self.getWatchers() + \
                           [x['login'] for x in self.getAllAssigned()] + \
                           [x['login'] for x in self.getAllPreAssigned()] + \
                           self.getAssociates():
                        self.watchers[0] += [(item['login'])]
                        self._assignLocalRoles(item['login'],
                                               'watcher',
                                               context)
                        self.setSubstitutesFor(item['login'],
                                               0,
                                               'watcher',
                                               context,
                                               with_local_roles=1)

    security.declareProtected(ModifyPortalContent, "doAssign")
    def doAssign(self, context=None, comment=''):
        """Assign/confirm_assign transition
        """

        self._p_changed = 1

        current_level = self.getCurrentLevel()
        wftool = self.portal_workflow

        if context is not None:
            review_history = wftool.getInfoFor(context, 'review_history')
            review_history.reverse()
            review_state = review_history[0]['review_state']
            if review_state == "waiting":
                try:
                    wftool.doActionFor(context,
                                       "assign",
                                       wf_id=self._getWorkflowId(),
                                       comment=comment)
                except (AttributeError, WorkflowException):
                    pass
            else:
                try:
                    wftool.doActionFor(context,
                                       "confirm_assign",
                                       wf_id=self._getWorkflowId(),
                                       comment=comment)
                except (AttributeError, WorkflowException):
                    pass

        # Updating pre-assigned -> assigned
        # Flushing next level
        self._updateLocalRoles(context)
        tmp = self.pilots.get(current_level+1, []) + \
              self.pre_pilots.get(current_level+1, [])
        self.pilots[current_level+1] = []
        self.pre_pilots[current_level+1] = []
        for id in tmp:
            if not self.isPilotSubstituteHere(id, current_level+1) and \
               not self.isPrePilotSubstituteHere(id, current_level+1):
                self.pilots[current_level+1] += [id]

        # Adding pilot to the list of pilots on the references documents
        for member in self.getPilots(level=current_level+1):
            # Metadata
            context.doAddPilot(member)
            # Substitutes
            self.setSubstitutesFor(member,
                                   current_level+1,
                                   'pilot',
                                   context)

        # Changing the current level
        self.current_level += 1

    security.declareProtected(ModifyPortalContent, "doInitializeReply")
    def doInitializeReply(self, context=None):
        """Initialize_reply transition

          - Follow the transition
          - Create the out going mail document
          - Associates become watchers -> watchers nothing
        """

        self._p_changed = 1

        if context is not None:
            wftool = context.portal_workflow

            #
            # Following the transition
            #

            try:
                wftool.doActionFor(context,
                                   'initialize_reply',
                                   wf_id=self._getWorkflowId(),)
            except (AttributeError, WorkflowException):
                pass

            #
            # Creation of the outgoing mail reference document
            #

            type_name = getattr(aq_base(self), 'outgoing_doc_type', 'mail_document')
            outgoing_docid = context.getOutGoingMailTypeId()
            outgoing_doc = None
            if not getattr(context, outgoing_docid, 0):
                context.invokeFactory(type_name, outgoing_docid)
                outgoing_doc = getattr(context, outgoing_docid)

            #
            # SRS : associates -> watchers
            #       watchers   -> nothing
            #

            #for watcher in self.getWatchers():
            #    self._removeLocalRoles(watcher, context)
            #self.watchers[0] = self.associates[0]
            #self.associates[0] = []
            #for ex_associate in self.getWatchers():
            #    self._removeLocalRoles(ex_associate, context)
            #    self._assignLocalRoles(ex_associate, 'watcher', context)

            return outgoing_doc

    security.declareProtected(ModifyPortalContent, "doReject")
    def doReject(self, context=None, comment=''):
        """Reject transition
        """

        self._p_changed = 1

        wftool = self.portal_workflow
        current_level = self.getCurrentLevel()

        if context is not None:

            try:
                wftool.doActionFor(context,
                                   'reject',
                                   wf_id=self._getWorkflowId(),
                                   comment=comment)
            except (AttributeError, WorkflowException):
                pass

            #
            # Removing local roles for the current pilots
            # 1 - Removing the LRs to the current pilots
            # 2 - Giving back LRs to the pilots above
            #

            for id in self.pilots[current_level]:
                self._removeLocalRoles(id, context)

            for pilot in self.getPilots(level=current_level-1):
                self._removeLocalRoles(pilot, context)
                self._assignLocalRoles(pilot, 'pilote', context)

            #
            # The current pilots are gonna be pre-pilots now.
            # Pilots->pre_pilots
            # Substitutes management
            # Decrementing current_level
            #

            pre_pilots = self.pre_pilots.get(current_level, [])
            for pilot in self.pilots[current_level]:
                if not self.isPilotSubstituteHere(pilot, current_level):
                    pre_pilots += [pilot]
            self.pre_pilots[current_level] = pre_pilots

            for pre_pilot in self.getPrePilots(level=current_level):
                self.setSubstitutesFor(pre_pilot,
                                       current_level,
                                       'pre_pilot',
                                       context)

            self.pilots[current_level] = []
            self.current_level = self.current_level - 1

    security.declareProtected(ModifyPortalContent, 'doClose')
    def doClose(self, context=None, comment=''):
        """Close transition

        Changing the mailfolder state to a 'closed' one.
        As well, locking the sub objects to avoid them being modified.
        Notice, delay is frozen after that.
        """

        self._p_changed = 1
        #Insert for publication by Cedric Portejoie (Capgemini) 2004/06/26
        try:
            mtool=self.portal_membership
            member_demand=str(mtool.getAuthenticatedMember())
        except:
            member_demand=''
        # End
        wftool = self.portal_workflow
        if context is not None:
            try:
                wftool.doActionFor(context,
                                   'close',
                                   wf_id=self._getWorkflowId(),
                                   comment=comment)
            except (AttributeError, WorkflowException):
                pass

            #
            # Publishing the outgoing_document within the pre-defined sections
            #

            document = getattr(context, context.getOutGoingMailTypeId(), None)

            if document is not None:
                # Set the publishing to and from dates to the same as the folder:
                doc_content = document.getContent()
                if self.date_available:
                    doc_content.date_available = self.date_available
                if self.date_issued:
                    doc_content.date_issued = self.date_issued 
                    allowed_transitions = wftool.getAllowedPublishingTransitions(
                        document)

                # XXX hack du jeudi
                allowed_transitions = [x for x in allowed_transitions
                                       if x != 'publish']

                for transition in allowed_transitions:
                    # If publishing environnement
                    rpaths = getattr(self, 'publication_sections', None)
                    if rpaths:
                        if isinstance(rpaths, StringType):
                            rpaths = (rpaths,)
                        for rpath in rpaths:
                            try:
                                LOG("MAILFOLDER PUBLICATION",
                                    DEBUG,
                                    rpaths,
                                    transition)
                                wftool.doActionFor(
                                    document,
                                    'copy_submit',
                                    dest_container=rpath,
                                    initial_transition=transition,
                                    comment=comment,
                                    demandeur=member_demand)
                            except WorkflowException:

                                #
                                # Here cause of the SRS.  Maybe one of the
                                # sections choosen in the metadata can't accept
                                # publication anymore....
                                #

                                pass

                #
                # Sub-objects are locked
                #

                for sub in context.objectValues():
                    try:
                        self.portal_workflow.doActionFor(sub,
                                                         'lock',
                                                         comment=comment)
                    except (AttributeError, WorkflowException):
                        pass
        #
        # Delay's not running anymore
        # Mailfolder stuffs
        #

        if getattr(self, 'freezeDelay', None):
            self.freezeDelay()

    security.declareProtected(ModifyPortalContent, 'doUnClose')
    def doUnClose(self, context=None, comment=''):
        """Unclose transition

        We gotta check the former state before the closing and then
        going back to this one through the corresponding transition.
        As well, the sub-contents will be unlocked.
        Notice, delay is not running anymore after that.
        """

        self._p_changed = 1

        wftool = self.portal_workflow

        if context is not None:

            #
            # Finding the transition we gotta follow depending on the former
            # state where we were before closing the mailfolder.
            #

            last_state_before_close = self._getFormerState(context)

            if last_state_before_close == "waiting":
                transition2follow = "unclose2waiting"
            elif last_state_before_close == "assigning":
                transition2follow = "unclose2assigning"
            elif last_state_before_close == "writing":
                transition2follow = "unclose2writing"
            elif last_state_before_close == "validating":
                transition2follow = "unclose2validating"
            else:
                LOG("MisillWorkflowEngine : WorkflowManager.unClose()",
                    INFO,
                    "Problem while trying to find the former transition for \
                    the mailfolder : ",
                    context.absolute_url(),
                    "Impossible to re-open it")

            #
            # Then following this one.
            #

            try:
                wftool.doActionFor(context,
                                   transition2follow,
                                   wf_id=self._getWorkflowId(),
                                   comment=comment)
            except (AttributeError, WorkflowException):
                pass

            #
            # Sub-objects will be unlocked
            #

            for sub in context.objectValues():
                try:
                    self.portal_workflow.doActionFor(sub,
                                                     'modification',
                                                     comment=comment)
                except (AttributeError, WorkflowException):
                    pass

        #
        # Delay running again
        # Mailfolder stuffs
        #

        if getattr(self, 'unFreezeDelay', None):
            self.unFreezeDelay()

    security.declareProtected(ModifyPortalContent, "doValidate")
    def doValidate(self, context=None, comment=''):
        """Validate transition

          1 - Call the validate transition
          2 - Current level pilots become associates
          3 - Decrementing the current level

          Notice that if we are on the top of the hierarchy we'll close
          the mail folder and it will be the end of the process.
          (Meaning no more pilots to validate)
        """

        self._p_changed = 1

        wftool = self.portal_workflow
        current_level = self.getCurrentLevel()

        if context is not None:
            review_state = wftool.getInfoFor(context,
                                             'review_state',
                                             wf_id=self._getWorkflowId())
            if self.getPilots(level='all').has_key(current_level-1):
                if self.getPilots(level=current_level-1) != []:
                    try:
                        wftool.doActionFor(context,
                                           'validate',
                                           wf_id=self._getWorkflowId(),
                                           comment=comment)
                    except (AttributeError, WorkflowException):
                        pass

                    #
                    # If this is the first validation step then the associates
                    # become watchers
                    #

                    former_state = self._getFormerState(context)
                    if former_state == 'writing':
                        for associate in self.getAssociates():
                            self._removeLocalRoles(associate, context)
                            if not self.isAssociateSubstituteHere(
                                associate,
                                current_level):
                                self._assignLocalRoles(associate,
                                                       'watcher',
                                                       context)
                                self.watchers[0] += [associate]
                                self.setSubstitutesFor(associate,
                                                       0,
                                                       'watcher',
                                                       context)
                        # Cleaning the assignment list
                        self.associates[0] = []

                    #
                    # We keep track of the current pilots for being able to
                    # reject the assignement on them again.  But they won't
                    # appear as pre-pilots but associates and they'll have
                    # associate's rights too.  As well, substitutes management
                    # in here since we are moving people to the pre-assigned
                    # list.
                    #

                    self.pre_pilots[current_level] = self.pre_pilots.get(
                        current_level,
                        [])
                    for pilot in self.pilots.get(current_level, []):
                        if not self.isPilotSubstituteHere(pilot,
                                                          current_level):
                            self.pre_pilots[current_level] += [pilot]
                            self.setSubstitutesFor(pilot,
                                                   current_level,
                                                   'pre_pilot',
                                                   context)

                    #
                    # We gotta change pilote -> associate except if
                    # If the guy is pilote at another level
                    #

                    current_pilots = self.getPilots(level=current_level)
                    del self.pilots[current_level]

                    all_pilots = [x['login'] for x in self.getAllAssigned()]

                    for pilot in current_pilots:
                        if pilot not in all_pilots:
                            self._removeLocalRoles(pilot, context)
                            self._assignLocalRoles(pilot,
                                                   'associate',
                                                   context)
                            if not self.isPilotSubstituteHere(pilot,
                                                              current_level):

                                #
                                # Substitutes will be added after through the
                                # substitutes API
                                #

                                self.associates[0] += [pilot]
                                self.setSubstitutesFor(pilot,
                                                       0,
                                                       'associate',
                                                       context)

                    #
                    # Now assigning rights to the future pilots at above level.
                    #

                    for pilot in self.getPilots(
                        level=self.getCurrentLevel()-1):
                        self._assignLocalRoles(pilot,
                                               'pilote',
                                               context)

                    #
                    # Sub objects  may be locked
                    # Thus Let's try to unlock them
                    # Could be a problem while validating
                    #

                    outgoing_docid = context.getOutGoingMailTypeId()
                    outgoing_doc = None
                    if getattr(context, outgoing_docid, 0):
                        outgoing_doc = getattr(context, outgoing_docid)
                        try:
                            self.portal_workflow.doActionFor(outgoing_doc,
                                                             'modification',
                                                             comment=comment)
                        except (AttributeError, WorkflowException):
                            pass

            else:

                #
                # Here we are on the top then we close the mail
                # Finnaly end of the process.
                #

                try:
                    wftool.doActionFor(context,
                                       'close',
                                       wf_id=self._getWorkflowId(),
                                       comment=comment)
                except (AttributeError, WorkflowException):
                    pass

            self.current_level -= 1

    security.declareProtected(ModifyPortalContent, "doRejectValidate")
    def doRejectValidate(self, context=None, comment=''):
        """Reject Validate transition

         - First, following the transition
         - Then, coping with assignment lists.
         - Then, local roles
         - Fixing associates list.
         - Incrementing level (going down again.)
        """

        self._p_changed = 1

        wftool = self.portal_workflow
        current_level = self.getCurrentLevel()

        if context is not None:
            try:
                wftool.doActionFor(context,
                                   'reject_validate',
                                   wf_id=self._getWorkflowId(),
                                   comment=comment)
            except (AttributeError, WorkflowException):
                pass

            #
            # Giving pilot rights at the level below.
            # Hack with pilots/pre_pilots in here since it's a sort of
            # unusual case
            # We can either have pilots or pre-pilots candidates.
            #

            new_next_level_pilots = self.getPilots(level=current_level+1)
            for pre_pilot in self.getPrePilots(level=current_level+1):
                new_next_level_pilots += [pre_pilot]
            self.pre_pilots[current_level+1] = []
            self.pilots[current_level+1] = []

            for pilot in new_next_level_pilots:
                if not self.isPilotSubstituteHere(pilot, current_level+1) and \
                   not self.isPrePilotSubstituteHere(pilot, current_level+1):
                    self.pilots[current_level+1] += [pilot]

            for pilot in self.getPilots(level=current_level+1):

                #
                # Changing local roles
                #

                self._removeLocalRoles(pilot, context)
                self._assignLocalRoles(pilot, 'pilote', context)

                #
                # Fixing associates (the pilots who became associates
                # the step before.
                #

                substitutes = self.getSubstitutesFor(pilot)
                new_associates = []
                for associate in self.getAssociates():
                    if associate != pilot and \
                           associate not in substitutes:
                        new_associates.append(associate)
                    else:
                        self.removeSubstitutesFor(pilot,
                                                  0,
                                                  'associate',
                                                  context)
                self.associates[0] = new_associates
                self.setSubstitutesFor(pilot,
                                       current_level+1,
                                       'pilot',
                                       context)

            self.current_level += 1

    security.declareProtected(ModifyPortalContent, 'doModify')
    def doModify(self, context=None, comment=''):
        """Modify transition

        Just for history.
        """

        if context is not None:
            wftool = context.portal_workflow
            try:
                wftool.doActionFor(context,
                                   'modify',
                                   wf_id=self._getWorkflowId(),
                                   comment=comment)
            except (AttributeError, WorkflowException):
                pass

    ##############################################################
    ##############################################################

    security.declarePublic('getClosingDate')
    def getClosingDate(self, context=None):
        """Return the closing date if the content is in a
        closed state

        If not closed then return None
        """

        if context is not None:
            review_state = self.portal_workflow.getInfoFor(context,
                                                           'review_state')
            if review_state == 'closed':
                review_history = self.portal_workflow.getInfoFor(
                    context,
                    'review_history')
                review_history.reverse()
                for state in review_history:
                    if state['action'] == 'close':
                        return state['time']
        return None

    def filterExistingUsers(self, list):
        #import pdb;pdb.set_trace()
        dirs = getToolByName(self, 'portal_directories')
        memberdir = dirs.members
        groupdir = dirs.groups
        res = []
        for each in list:
            if ':' in each:
                name = each.split(':')[1]
                if groupdir.hasEntry(name):
                    res.append(each)
            else:
                if memberdir.hasEntry(each):
                    res.append(each)
        return res            

InitializeClass(WorkflowManager)
